@@ -9,4 +9,4 @@ from .minimize import Optimizer
 from .hessian_approximation import SR1, BFGS, DFP, HessianApproximation
 from .logging import logger
 from .version import __version__
-from .constants import Options, SubSpaceDim
+from .constants import Options, SubSpaceDim, StepBackStrategy
