@@ -10,9 +10,9 @@ import numpy as np
 import logging
 from numpy.linalg import norm
 from scipy.sparse import csc_matrix
-from .trust_region import trust_region_reflective
+from .trust_region import trust_region_reflective, Step
 from .hessian_approximation import HessianApproximation
-from .constants import Options, DEFAULT_OPTIONS
+from .constants import Options, StepBackStrategy, ExitFlag, DEFAULT_OPTIONS
 from .logging import logger
 
 from typing import Callable, Dict, Optional, Tuple
@@ -22,20 +22,25 @@ class Optimizer:
     """
     Performs optimization
 
-    :ivar fun: objective function
-    :ivar funargs: keyword arguments that are passed to the function
-    :ivar lb: lower optimization boundaries
-    :ivar ub: upper optimization boundaries
-    :ivar options: options that configure convergence checks
-    :ivar delta: trust region radius
-    :ivar x: current optimization variables
-    :ivar fval: objective function value at x
-    :ivar grad: objective function gradient at x
-    :ivar hess: objective function Hessian (approximation) at x
-    :ivar hessian_update: object that performs hessian updates
-    :ivar starttime: time at which optimization was started
-    :ivar iteration: current iteration
-    :ivar converged: flag indicating whether optimization has converged
+    :ivar fun: Objective function
+    :ivar funargs: Keyword arguments that are passed to the function
+    :ivar lb: Lower optimization boundaries
+    :ivar ub: Upper optimization boundaries
+    :ivar options: Options that configure convergence checks
+    :ivar delta_iter: Trust region radius that was used for the current step
+    :ivar delta: Updated trust region radius
+    :ivar x: Current optimization variables
+    :ivar fval: Objective function value at x
+    :ivar grad: Objective function gradient at x
+    :ivar x_min: Optimal optimization variables
+    :ivar fval_min: Objective function value at x_min
+    :ivar grad_min: Objective function gradient at x_min
+    :ivar hess: Objective function Hessian (approximation) at x
+    :ivar hessian_update: Object that performs hessian updates
+    :ivar starttime: Time at which optimization was started
+    :ivar iteration: Current iteration
+    :ivar converged: Flag indicating whether optimization has converged
+    :ivar exitflag:
     """
     def __init__(self, fun: Callable,
                  ub: np.ndarray,
@@ -88,20 +93,35 @@ class Optimizer:
 
         self.options: Dict = options
 
-        self.delta = 10
+        self.delta: float = self.get_option(Options.DELTA_INIT)
+        self.delta_iter: float = self.delta
+
+        self.tr_ratio: float = 1
 
         self.x: np.ndarray = np.empty(ub.shape)
-        self.fval: float = np.nan
+        self.fval: float = np.inf
         self.grad: np.ndarray = np.empty(ub.shape)
         self.hess: np.ndarray = np.empty((ub.shape[0], ub.shape[0]))
+        self.x_min = self.x
+        self.fval_min = self.fval
+        self.grad_min = self.grad
 
         self.hessian_update: HessianApproximation = hessian_update
 
         self.starttime: float = np.nan
         self.iteration: int = 0
         self.converged: bool = False
+        self.exitflag: ExitFlag = ExitFlag.DID_NOT_RUN
         logger.setLevel(verbose)
 
+    def _reset(self):
+        self.starttime = time.time()
+        self.iteration = 0
+        self.converged: bool = False
+        self.delta: float = self.get_option(Options.DELTA_INIT)
+        self.delta_iter: float = self.delta
+        self.fval_min = np.inf
+
     def minimize(self, x0: np.ndarray):
         """
         Minimize the objective function the interior trust-region reflective
@@ -126,73 +146,160 @@ class Optimizer:
             grad: final gradient,
             hess: final Hessian (approximation)
         """
-        self.starttime = time.time()
-        self.iteration = 0
+        self._reset()
 
         self.x = np.array(x0).copy()
+        if self.x.ndim > 1:
+            raise ValueError('x0 must be a vector with x.ndim == 1!')
         self.make_non_degenerate()
         self.check_in_bounds()
         if self.hessian_update is None:
             self.fval, self.grad, self.hess = self.fun(self.x, **self.funargs)
         else:
             self.fval, self.grad = self.fun(self.x, **self.funargs)
+            self.hessian_update.init_mat(len(self.x))
             self.hess = self.hessian_update.get_mat()
 
-        self.check_finite()
+        if not np.isscalar(self.fval):
+            raise ValueError('Provided objective function must return a '
+                             'scalar!')
+        if not self.grad.ndim == 1:
+            raise ValueError('Provided objective function must return a '
+                             'gradient vector with x.ndim == 1, was '
+                             f'{self.grad.ndim}!')
+        if not len(self.grad) == len(self.x):
+            raise ValueError('Provided objective function must return a '
+                             'gradient vector of the same shape as x, '
+                             f'x has {len(self.x)} entries but gradient has '
+                             f'{len(self.grad)}!')
 
-        tr_space = None
+        if not len(self.grad) == len(self.x):
+            raise ValueError('Provided objective function must return a '
+                             'gradient vector of the same shape as x, '
+                             f'x has {len(self.x)} entries but gradient has '
+                             f'{len(self.grad)}!')
+
+        # hessian approximation would error on these earlier
+        if not self.hess.ndim == 2:
+            raise ValueError('Provided objective function must return a '
+                             'Hessian matrix with x.ndim == 2, was '
+                             f'{self.hess.ndim}!')
+
+        if not self.hess.shape[0] == self.hess.shape[1]:
+            raise ValueError('Provided objective function must return a '
+                             'square Hessian matrix!')
+
+        if not self.hess.shape[0] == len(self.x):
+            raise ValueError('Provided objective function must return a '
+                             'square Hessian matrix with same dimension'
+                             f'x has {len(self.x)} entries but Hessian has '
+                             f'{self.hess.shape[0]}!')
+
+        self.track_minimum(self.x, self.fval, self.grad)
+        self.log_header()
+        self.log_step_initial()
+
+        self.check_finite()
 
         self.converged = False
 
         while self.check_continue():
             self.iteration += 1
+            self.delta_iter = self.delta
 
             v, dv = self.get_affine_scaling()
 
             scaling = csc_matrix(np.diag(np.sqrt(np.abs(v))))
-            theta = max(.95, 1 - norm(v * self.grad, np.inf))
+            theta = max(self.get_option(Options.THETA_MAX),
+                        1 - norm(v * self.grad, np.inf))
 
-            step_x, step_sx, qppred, tr_space, step_type = \
+            step = \
                 trust_region_reflective(
-                    self.x, self.grad, self.hess, scaling, tr_space,
+                    self.x, self.grad, self.hess, scaling,
                     self.delta, dv, theta, self.lb, self.ub,
-                    self.get_option(Options.SUBSPACE_DIM)
+                    self.get_option(Options.SUBSPACE_DIM),
+                    self.get_option(Options.STEPBACK_STRAT)
                 )
 
-            x_new = self.x + step_x
+            x_new = self.x + step.s + step.s0
 
             if self.hessian_update is None:
                 fval_new, grad_new, hess_new = self.fun(x_new, **self.funargs)
             else:
                 fval_new, grad_new = self.fun(x_new, **self.funargs)
+                hess_new = None
 
-            accepted = self.update_tr_radius(
-                fval_new, grad_new, step_sx, dv, qppred
-            )
+            accepted = self.update_tr_radius(fval_new, grad_new, step, dv)
 
-            if (self.iteration - 1) % 10 == 0:
+            if self.iteration % 10 == 0:
                 self.log_header()
-            self.log_step(accepted, step_type, norm(step_x))
+            self.log_step(accepted, step, theta, fval_new)
             self.check_convergence(fval_new, x_new, grad_new)
 
-            if accepted:
-                if self.hessian_update is not None:
-                    self.hessian_update.update(step_x, grad_new - self.grad)
-                    self.hess = self.hessian_update.get_mat()
-                else:
-                    self.hess = hess_new
-                self.check_in_bounds(x_new)
-                self.fval = fval_new
-                self.x = x_new
-                self.grad = grad_new
-                self.check_finite()
-                self.make_non_degenerate()
+            # track minimum independently of whether we accept the step or not
+            self.track_minimum(x_new, fval_new, grad_new)
 
-                tr_space = None
+            if accepted:
+                self.update(step, x_new, fval_new, grad_new, hess_new)
 
         return self.fval, self.x, self.grad, self.hess
 
-    def update_tr_radius(self, fval, grad, step_sx, dv, qppred) -> bool:
+    def track_minimum(self,
+                      x_new: np.ndarray,
+                      fval_new: float,
+                      grad_new: np.ndarray) -> None:
+        """
+        Function that tracks the optimization variables that have minimal
+        function value independent of whether the step is accepted or not.
+
+        :param x_new:
+        :param fval_new:
+        :param grad_new:
+        :return:
+        """
+        if np.isfinite(fval_new) and fval_new < self.fval_min:
+            self.x_min = x_new
+            self.fval_min = fval_new
+            self.grad_min = grad_new
+
+    def update(self,
+               step: Step,
+               x_new: np.ndarray,
+               fval_new: float,
+               grad_new: np.ndarray,
+               hess_new: Optional[np.ndarray] = None) -> None:
+        """
+        Update self according to employed step
+
+        :param step:
+            Employed step
+        :param x_new:
+            New optimization variable values
+        :param fval_new:
+            Objective function value at x_new
+        :param grad_new:
+            Objective function gradient at x_new
+        :param hess_new:
+            (Approximate) objective function Hessian at x_new
+        """
+        if self.hessian_update is not None:
+            self.hessian_update.update(step.s + step.s0,
+                                       grad_new - self.grad)
+            self.hess = self.hessian_update.get_mat()
+        else:
+            self.hess = hess_new
+        self.check_in_bounds(x_new)
+        self.fval = fval_new
+        self.x = x_new
+        self.grad = grad_new
+        self.check_finite()
+        self.make_non_degenerate()
+
+    def update_tr_radius(self,
+                         fval: float,
+                         grad: np.ndarray,
+                         step: Step,
+                         dv: np.ndarray) -> bool:
         """
         Update the trust region radius
 
@@ -200,32 +307,39 @@ class Optimizer:
             new function value if step defined by step_sx is taken
         :param grad:
             new gradient value if step defined by step_sx is taken
-        :param step_sx:
-            proposed scaled step
+        :param step:
+            step
         :param dv:
             derivative of scaling vector v wrt x
-        :param qppred:
-            predicted objective function value according to the quadratic
-            approximation
 
         :return:
             flag indicating whether the proposed step should be accepted
         """
-        nsx = norm(step_sx)
+        stepsx = step.ss + step.ss0
+        nsx = norm(stepsx)
         if not np.isfinite(fval):
-            self.delta = np.min([self.delta / 4, nsx / 4])
+            self.delta = np.nanmin([
+                self.delta * self.get_option(Options.GAMMA1),
+                nsx / 4
+            ])
             return False
         else:
-            qpval = 0.5 * step_sx.dot(dv * np.abs(grad) * step_sx)
-            ratio = (fval + qpval - self.fval) / qppred
+            qpval = 0.5 * stepsx.dot(dv * np.abs(grad) * stepsx)
+            self.tr_ratio = (fval + qpval - self.fval) / step.qpval
 
             # values as proposed in algorithm 4.1 in Nocedal & Wright
-            if ratio >= 0.75 and nsx > self.delta * 0.9:
-                self.delta = 2 * self.delta
-            elif ratio <= .25 or nsx < self.delta * 0.9 \
-                    or fval > self.fval * 1.1:
-                self.delta = np.min([self.delta / 4, nsx / 4])
-            return ratio >= .25 and fval < self.fval * 1.1
+            if self.tr_ratio >= self.get_option(Options.ETA) \
+                    and nsx > self.delta * 0.9:
+                # increase radius
+                self.delta = self.get_option(Options.GAMMA2) * self.delta
+            elif self.tr_ratio <= self.get_option(Options.MU) \
+                    or nsx < self.delta * 0.9:
+                # decrease radius
+                self.delta = np.nanmin([
+                    self.delta * self.get_option(Options.GAMMA1),
+                    nsx / 4
+                ])
+            return self.tr_ratio >= self.get_option(Options.MU)
 
     def check_convergence(self, fval, x, grad) -> None:
         """
@@ -249,6 +363,7 @@ class Optimizer:
         gnorm = norm(grad)
 
         if np.isclose(fval, self.fval, atol=fatol, rtol=frtol):
+            self.exitflag = ExitFlag.FTOL
             logger.info(
                 'Stopping as function difference '
                 f'{np.abs(self.fval - fval)} was smaller than specified '
@@ -257,6 +372,7 @@ class Optimizer:
             converged = True
 
         elif np.isclose(x, self.x, atol=xatol, rtol=xrtol).all():
+            self.exitflag = ExitFlag.XTOL
             logger.info(
                 'Stopping as step was smaller than specified tolerances ('
                 f'atol={xatol:.2E}, rtol={xrtol:.2E})'
@@ -264,6 +380,7 @@ class Optimizer:
             converged = True
 
         elif gnorm <= gatol:
+            self.exitflag = ExitFlag.GTOL
             logger.info(
                 'Stopping as gradient norm satisfies absolute convergence '
                 f'criteria: {gnorm:.2E} < {gatol:.2E}'
@@ -271,6 +388,7 @@ class Optimizer:
             converged = True
 
         elif gnorm <= grtol * self.fval:
+            self.exitflag = ExitFlag.GTOL
             logger.info(
                 'Stopping as gradient norm satisfies relative convergence '
                 f'criteria: {gnorm:.2E} < {grtol:.2E} * {self.fval:.2E}'
@@ -293,6 +411,7 @@ class Optimizer:
 
         maxiter = self.get_option(Options.MAXITER)
         if self.iteration > maxiter:
+            self.exitflag = ExitFlag.MAXITER
             logger.error(
                 f'Stopping as maximum number of iterations {maxiter} was '
                 f'exceeded.'
@@ -304,6 +423,7 @@ class Optimizer:
         time_remaining = maxtime - time_elapsed
         avg_iter_time = time_elapsed/(self.iteration + (self.iteration == 0))
         if time_remaining < avg_iter_time:
+            self.exitflag = ExitFlag.MAXTIME
             logger.error(
                 f'Stopping as maximum runtime {maxtime} is expected to be '
                 f'exceeded in the next iteration.'
@@ -311,6 +431,7 @@ class Optimizer:
             return False
 
         if self.delta < np.spacing(1):
+            self.exitflag = ExitFlag.SMALL_DELTA
             logger.error(
                 'Stopping as trust region radius is smaller that machine '
                 'precision.'
@@ -356,38 +477,82 @@ class Optimizer:
         dv[bounded] = 1
         return v, dv
 
-    def log_step(self, accepted: bool, steptype: str, normdx: float):
+    def log_step(self, accepted: bool, step: Step, theta: float, fval: float):
         """
         Prints diagnostic information about the current step to the log
 
         :param accepted:
             flag indicating whether the current step was accepted
-        :param steptype:
-            identifier how the current step was computed
-        :param normdx:
-            norm of the current step
+        :param step:
+            proposal step
+        :param theta:
+            value of the theta parameter
+        :param fval:
+            new fval if step is accepted
         """
+        normdx = norm(step.s + step.s0)
+
         iterspaces = max(len(str(self.get_option(Options.MAXITER))), 5) - \
             len(str(self.iteration))
-        steptypespaces = 4 - len(steptype)
+        steptypespaces = 4 - len(step.type)
+        if self.get_option(Options.STEPBACK_STRAT) == StepBackStrategy.REFLECT:
+            count = step.reflection_count
+        else:
+            count = step.truncation_count
+        stepbackspaces = 4 - len(str(count))
+        if np.isnan(fval):
+            fval = self.fval
         logger.info(f'{" " * iterspaces}{self.iteration}'
-                    f' | {self.fval:.4E}'
-                    f' | {self.delta:.2E}'
+                    f' | {fval if accepted else self.fval:.3E}'
+                    f' | {(fval - self.fval)*accepted:+.2E}'
+                    f' | {step.qpval:+.2E}'
+                    f' | {self.tr_ratio:+.2E}'
+                    f' | {self.delta_iter:.2E}'
                     f' | {norm(self.grad):.2E}'
                     f' | {normdx:.2E}'
-                    f' | {steptype}{" " * steptypespaces}'
+                    f' | {theta:.2E}'
+                    f' | {step.alpha:.2E}'
+                    f' | {step.type}{" " * steptypespaces}'
+                    f' | {" " * stepbackspaces}{count}'
                     f' | {int(accepted)}')
 
+    def log_step_initial(self):
+        """
+        Prints diagnostic information about the initial step to the log
+        """
+
+        iterspaces = max(len(str(self.get_option(Options.MAXITER))), 5) - \
+            len(str(self.iteration))
+        logger.info(f'{" " * iterspaces}{self.iteration}'
+                    f' | {self.fval:.3E}'
+                    f' |    NaN   '
+                    f' |    NaN   '
+                    f' |    NaN   '
+                    f' | {self.delta:.2E}'
+                    f' | {norm(self.grad):.2E}'
+                    f' |   NaN   '
+                    f' |   NaN   '
+                    f' |   NaN   '
+                    f' | NaN '
+                    f' | NaN '
+                    f' | {int(np.isfinite(self.fval))}')
+
     def log_header(self):
         """
         Prints the header for diagnostic information, should complement
         :py:func:`Optimizer.log_step`.
         """
         iterspaces = len(str(self.get_option(Options.MAXITER))) - 5
+
+        if self.get_option(Options.STEPBACK_STRAT) == StepBackStrategy.REFLECT:
+            countheader = 'refl'
+        else:
+            countheader = 'trun'
+
         logger.info(f'{" " * iterspaces} iter '
-                    f'|    fval    |  delta   | ||step|| |  ||g||   '
-                    f'| step | '
-                    f'accept')
+                    f'|   fval    | fval diff | pred diff | tr ratio  '
+                    f'|  delta   |  ||g||   | ||step|| |  theta   |  alpha   '
+                    f'| step | {countheader} | accept')
 
     def check_finite(self):
         """
@@ -404,16 +569,19 @@ class Optimizer:
             pointstr = f'at iteration {self.iteration}.'
 
         if not np.isfinite(self.fval):
+            self.exitflag = ExitFlag.NOT_FINITE
             raise RuntimeError(f'Encountered non-finite function {self.fval} '
                                f'value {pointstr}')
 
         if not np.isfinite(self.grad).all():
+            self.exitflag = ExitFlag.NOT_FINITE
             ix = np.where(np.logical_not(np.isfinite(self.grad)))
             raise RuntimeError('Encountered non-finite gradient entries'
                                f' {self.grad[ix]} for indices {ix} '
                                f'{pointstr}')
 
         if not np.isfinite(self.hess).all():
+            self.exitflag = ExitFlag.NOT_FINITE
             ix = np.where(np.logical_not(np.isfinite(self.hess)))
             raise RuntimeError('Encountered non-finite gradient hessian'
                                f' {self.hess[ix]} for indices {ix} '
@@ -440,8 +608,9 @@ class Optimizer:
                                    ['upper bounds', 'lower bounds']):
             diff = sign * (ref - x)
             if not np.all(diff <= 0):
-                ix = np.where(diff > 0)
-                raise RuntimeError(f'Exceeded upper bounds for indices {ix} by'
+                ix = np.where(diff > 0)[0]
+                self.exitflag = ExitFlag.EXCEEDED_BOUNDARY
+                raise RuntimeError(f'Exceeded {name} for indices {ix} by '
                                    f'{diff[ix]} {pointstr}')
 
     def get_option(self, option):
